#include <sys/queue.h>

#include <stdlib.h>
#include <string.h>

#include "spho/spho.h"
#include "spho/ctx.h"
#include "spho/scope.h"
#include "spho/bind.h"

static int scope_nom_lookup_str_local(struct spho_scope *, const char *,
    size_t, struct spho_nom **);
static int scope_nom_in_local(struct spho_scope *, struct spho_nom *);

void
spho_scope_init(struct spho_scope *sc, struct spho_ctx *ctx,
    struct spho_scope *p)
{
	sc->ctx = ctx;
	sc->parent = p;

	STAILQ_INIT(&sc->subs);

	SLIST_INIT(&sc->noms);
	sc->nom_cnt = 0;

	STAILQ_INIT(&sc->tps);

	sc->bind_loc = NULL;
}

void
spho_scope_cleanup(struct spho_scope *sc)
{
	struct spho_scope *sub;
	struct spho_nom *nom;
	struct spho_tp_alloc *tp_alloc;

	/* XXX is doable without recursive call, but cba */
	while (! STAILQ_EMPTY(&sc->subs)) {
		sub = STAILQ_FIRST(&sc->subs);
		STAILQ_REMOVE_HEAD(&sc->subs, entries);
		spho_scope_cleanup(sub);
		free(sub);
	}

	/* all type structs are allocated from scope */
	while (! STAILQ_EMPTY(&sc->tps)) {
		tp_alloc = STAILQ_FIRST(&sc->tps);
		STAILQ_REMOVE_HEAD(&sc->tps, entries);
		free(tp_alloc);
	}

	/* names are allocated in scope too */
	while (! SLIST_EMPTY(&sc->noms)) {
		nom = SLIST_FIRST(&sc->noms);
		SLIST_REMOVE_HEAD(&sc->noms, entries);
		free(nom);
	}

	sc->parent = NULL;

	if (sc->bind_loc != NULL)
		spho_prebind_free(sc->bind_loc);
	sc->bind_loc = NULL;
}

struct spho_scope *
spho_scope_global(struct spho_ctx *ctx)
{
	SPHO_PRECOND(ctx != NULL);

	return (&ctx->glob);
}


struct spho_scope *
spho_scope_create_subscope(struct spho_scope *sc)
{
	struct spho_scope *sub;

	if ((sub = malloc(sizeof(struct spho_scope))) == NULL) {
		SPHO_ERR(sc->ctx, SPHO_ERR_SYS);
		return (NULL);
	}

	spho_scope_init(sub, sc->ctx, sc);

	STAILQ_INSERT_TAIL(&sc->subs, sub, entries);

	return (sub);
}

void
spho_scope_free(struct spho_scope *sc)
{
	struct spho_scope *p;

	p = sc->parent;
	STAILQ_REMOVE(&p->subs, sc, spho_scope, entries);
	spho_scope_cleanup(sc);
	free(sc);
}

struct spho_nom *
spho_scope_nom_add(struct spho_scope *sc, const char *nomstr, size_t sz)
{
	size_t res;
	struct spho_nom *nom;

	nom = NULL;

	if (scope_nom_lookup_str_local(sc, nomstr, sz, &nom) == -1)
		return (NULL);

	if (nom != NULL) {
		SPHO_SC_ERR(sc, SPHO_ERR_NOM_INUSE);
		return (NULL);
	}

	if ((nom = malloc(sizeof(struct spho_nom))) == NULL) {
		SPHO_SC_ERR(sc, SPHO_ERR_SYS);
		return (NULL);
	}

	res = strlcpy(nom->s, nomstr, sizeof(nom->s));
	if (res >= sizeof(nom->s)) {
		SPHO_SC_ERR(sc, SPHO_ERR_TOOBIG);
		goto err;
	}

	SLIST_INSERT_HEAD(&sc->noms, nom, entries);
	sc->nom_cnt++;

	return (nom);
err:
	free(nom);
	return (NULL);
}


int
spho_scope_nom_lookup(struct spho_scope *sc, const char *nomstr, size_t sz,
    struct spho_nom **out)
{
	struct spho_nom *nom;

	nom = NULL;

	if (sz > sizeof(nom->s)) {
		SPHO_ERR(sc->ctx, SPHO_ERR_TOOBIG);
		return (-1);
	}

	*out = NULL;
	while (*out == NULL && sc != NULL) {
		SLIST_FOREACH(nom, &sc->noms, entries) {
			if (strncmp(nom->s, nomstr, sz) == 0) {
				*out = nom;
				break;
			}
		}
		sc = sc->parent;
	}

	return (0);
}

int
spho_scope_nom_lookup_strict(struct spho_scope *sc, const char *nomstr,
    size_t sz, struct spho_nom **out)
{
	if (spho_scope_nom_lookup(sc, nomstr, sz, out) == -1)
		return (-1);

	if (*out == NULL) {
		SPHO_SC_ERR(sc, SPHO_ERR_NOM_NOTINSCOPE);
		return (-1);
	}

	return (0);
}

int
spho_scope_prebind_init(struct spho_scope *sc)
{
	if ((sc->bind_loc = spho_prebind_create(sc)) == NULL)
		return (-1);

	return (0);
}

int
spho_scope_prebind_tp(struct spho_scope *sc, struct spho_nom *nom,
    struct spho_tp *tp)
{
	if (! scope_nom_in_local(sc, nom)) {
		SPHO_ERR(sc->ctx, SPHO_ERR_NOM_NOTINSCOPE);
		return (-1);
	}

	return (spho_prebind_tp(sc->bind_loc, nom, tp));
}

int
spho_scope_prebind_tp_op(struct spho_scope *sc, struct spho_nom *nom,
    struct spho_tp_op *tp_op)
{
	if (! scope_nom_in_local(sc, nom)) {
		SPHO_ERR(sc->ctx, SPHO_ERR_NOM_NOTINSCOPE);
		return (-1);
	}

	return (spho_prebind_tp_op(sc->bind_loc, nom, tp_op));
}

int
spho_scope_prebind_undef(struct spho_scope *sc, struct spho_nom *nom)
{
	if (! scope_nom_in_local(sc, nom)) {
		SPHO_ERR(sc->ctx, SPHO_ERR_NOM_NOTINSCOPE);
		return (-1);
	}

	return (spho_prebind_undef(sc->bind_loc, nom));
}

static int
scope_nom_lookup_str_local(struct spho_scope *sc, const char *nomstr,
    size_t sz, struct spho_nom **out)
{
	struct spho_nom *nom;

	nom = NULL;

	if (sz > sizeof(nom->s)) {
		SPHO_ERR(sc->ctx, SPHO_ERR_TOOBIG);
		return (-1);
	}

	*out = NULL;
	SLIST_FOREACH(nom, &sc->noms, entries) {
		if (strncmp(nom->s, nomstr, sz) == 0) {
			*out = nom;
			break;
		}
	}

	return (0);
}

static int
scope_nom_in_local(struct spho_scope *sc, struct spho_nom *nom)
{
	struct spho_nom *sc_nom;

	SLIST_FOREACH(sc_nom, &sc->noms, entries) {
		if (sc_nom == nom)
			return (1);
	}

	return (0);
}
